/*      XQ      Xiang Qi        */

#include <htext.h>
#include <na.h>
#include <fdata.h>

#include "stdio.h"

extern int mkik(), mkobj(), sasa(), sa_swap();
int xq_rscan(), xq_cscan(), xq_enprise(), rc_line();
extern int qt;

#define MAX_MOVES 125
#define B_SIZE  90
#define NR      10
#define NC      9
#define RIVER   4

#define FC_END   '.'    /* test for at least one legal move */
#define FC_LEGAL '-'    /* filter legal moves */
#define FC_CHECK '+'    /* legal moves. mark checks */

/* piece values. Swap by exclusive or with 8 */

#define PAWN       1
#define ADVISOR    2
#define ELEPHANT   3
#define CANNON     4
#define HORSE      5
#define ROOK       6
#define KING       7
#define B_KING    15

#define B_P      9
#define B_A     10
#define B_E     11
#define B_C     12
#define B_H     13
#define B_R     14
#define B_K     15

/* pieces in black palace */

#define A_BOX(X) (((X)==66)||((X)==68)||((X)==76)||((X)==84)||((X)==86))
#define K_BOX(X,Y) ((X)>2 && (X)<6 && (Y)>6)

#define I_SIGN(X)        (((X)>0) ? 1 : -1)
#define I_ABS(X)         (((X)>0) ? (X) : -(X))

#define XY_PUT(DST,X,Y) { *(DST)++ = (X); *(DST)++ = (Y); }

int     xq_dz[]= {1, NC, -1, -NC };

int xq_moves(dst, lo, hi)   /* Moves in Xiang qi position */
MATRIX  dst;            /* (nmoves x 2) table of from .. to pairs */
MATRIX  lo;             /* options string */
MATRIX  hi;             /* ASCII board. 90 characters */
{
int     xq_flip();
int    *map = INT_PTR(hi);
BYTE   *fmap = NULL;
int     fc = 0;
int     qb = 0;
int     rc = 0;
int     j, x, dx;
int     cx, rx;         /* row, column */
int     pwk, pbk;       /* Kings */
int     ny, y;          /* y-points */
int     hit;            /* hit by cannon */
int     cnt;
int    *z, *si;
int     k_file[B_SIZE]; /* illegal king moves */
int     xx[B_SIZE];     /* probe result of move */

if (B_SIZE != MSIZE(hi->ADDR)) return ERROR;
if (MTYPE(hi->ADDR) != TYPN) return ERROR;

if  ((NULL != lo) &&
    (NULL != lo->ADDR) &&
    (TYPT == MTYPE(lo->ADDR)) &&
    (0 < (j = MSIZE(lo->ADDR)))) {
        fmap = TEXT_PTR(lo);
        if ('B' == *fmap || 'R' ==  *fmap) {
            qb = ('B' == *fmap);
            fmap++;
            j--;
            }
        if (j > 0) fc = *fmap;
        }

if (qb) xq_flip(hi);

/* must have two kings */

pwk = pbk = 0;
for (x = 0; x < B_SIZE; x++) {
    k_file[x] = 0;
    if ( map[x] == KING) pwk = x;
    if ( map[x] == B_KING) pbk = x;
    }

if ( pwk * pbk == 0) return mkobj(dst, TYPN, 0, 0);

ny = xq_rscan(map, pbk, -NC);
for (x = pbk - NC; 0<ny--; x -= NC) k_file[x] = 1;

mkobj(dst, TYPN, 2*MAX_MOVES, 2);


for (z = INT_PTR(dst), rx = cx = x = 0; x < B_SIZE; cx++, x++) {
    if (cx == NC) { cx = 0; rx++; }
    switch (map[x]) {
    case PAWN:
        if (rx < NR-1) XY_PUT(z, x, x+NC);
        if (rx > RIVER && cx > 0) XY_PUT(z, x, x-1);
        if (rx > RIVER && cx < NC-1) XY_PUT(z, x, x+1);
        break;
    case ADVISOR:       /* up to four diagonal moves */
        if (rx > 0 && cx > 3) XY_PUT(z, x, x - NC - 1);
        if (rx > 0 && cx < 5) XY_PUT(z, x, x - NC + 1);
        if (rx < 2 && cx > 3) XY_PUT(z, x, x + NC - 1);
        if (rx < 2 && cx < 5) XY_PUT(z, x, x + NC + 1);
        break;
    case ELEPHANT:      /* up to four diagonal jumps */
        if (rx > 1 && cx > 1 && 0 == map[x-NC-1])
            XY_PUT(z, x, x - 2*NC - 2);
        if (rx > 1 && cx < NC-2 && 0 == map[x-NC+1])
            XY_PUT(z, x, x - 2*NC + 2);
        if (rx < RIVER && cx > 1 && 0 == map[x+NC-1])
            XY_PUT(z, x, x + 2*NC - 2);
        if (rx < RIVER && cx < NC-2 && 0 == map[x+NC+1])
            XY_PUT(z, x, x + 2*NC + 2);
        break;
    case CANNON:
        for (j = 0; j<4; j++) {
            dx = xq_dz[j];
            ny = xq_cscan(map, x, dx, &hit);
            for (y = x+dx; 0 < ny--; y += dx) XY_PUT(z, x, y);
            if (hit > 0) XY_PUT(z, x, x + dx * hit);
            }
        break;
    case HORSE: /* up to 8 moves */
        if (rx < NR-2 && 0 == map[x + NC]) {
            if (cx > 0) XY_PUT(z, x, x + 2*NC - 1);
            if (cx < NC-1) XY_PUT(z, x, x + 2*NC+1);
            }
        if (rx > 1 && 0 == map[x - NC]) {
            if (cx > 0) XY_PUT(z, x, x - 2*NC - 1);
            if (cx < NC-1) XY_PUT(z, x, x - 2*NC+1);
            }

        if (cx < NC-2 && 0 == map[x + 1]) {
            if (rx > 0) XY_PUT(z, x, x + 2 - NC);
            if (rx < NR-1) XY_PUT(z, x, x + 2 + NC);
            }

        if (cx > 1 && 0 == map[x - 1]) {
            if (rx > 0) XY_PUT(z, x, x - 2 - NC);
            if (rx < NR-1) XY_PUT(z, x, x - 2 + NC);
            }
        break;
    case ROOK: /* scan each direction */
        for (j = 0; j<4; j++) {
            dx = xq_dz[j];
            ny = xq_rscan(map, x, dx);
            for (y = x+dx; 0 < ny--; y += dx) XY_PUT(z, x, y);
            }
        break;
    case KING:  /* up to four moves, all in box */
        if (rx < 2 && 0 == k_file[y = x+NC]) XY_PUT(z, x, y);
        if (rx > 0 && 0 == k_file[y = x-NC]) XY_PUT(z, x, y);
        if (cx > 3 && 0 == k_file[y = x -1]) XY_PUT(z, x, y);
        if (cx < 5 && 0 == k_file[y = x +1]) XY_PUT(z, x, y);
    default: break;
        }
    }
/* now trim the destination */

for (z = INT_PTR(dst), ny = 0; ny < MAX_MOVES && z[0] != z[1]; ny++, z+=2);

if (fc == 0) {  /* all possible moves */
    if (mkobj (hi, TYPN, 2*ny, 2)) return ERROR;

    for (ny = 2*ny, z = INT_PTR(hi), si = INT_PTR(dst); 0 < ny--;) {
        x = *si++;
       *z++ = (qb) ? 89-x : x;    /* was it flipped */
        }
    return sa_swap(dst, hi);
    }

if (fc == '.') {   /* at least one move */
    rc = 1;     /* game over */
    for (si = INT_PTR(dst); rc == 1 && 0 < ny--; si+=2) {
        x = map[si[1]];
        if (x == 0 || x > 8) {  /* legal move */
            for (j = 0; j < B_SIZE; j++) xx[j] = map[j];
            xx[si[1]] = xx[si[0]];
            xx[si[0]] = 0;
/*          if (qt) xq_debug("xq.log", ny, xx);   */
            rc = (0 < xq_enprise(xx, ERROR));
            }
        }
    return mkik(dst, rc);       /* 0= OK, 1 = no legal move */
    }

if (fc == '-' || fc == '+') {   /* list all legal moves */
    for (cnt = ny, si = INT_PTR(dst); 0 < cnt--; si+=2) {
        j  = map[si[1]];
        if (j > 0 && j < 8) si[0] = ERROR;
        else {
            for (j = 0; j < B_SIZE; j++) xx[j] = map[j];
            xx[si[1]] = xx[si[0]];
            xx[si[0]] = 0;
            if (0 < xq_enprise(&xx[0], ERROR)) si[0] = ERROR;
            }
        }
    for (cnt = 0, j = ny, si = INT_PTR(dst); 0 < j--; si+=2)
        cnt += (si[0] != ERROR);        /* table rows */
    }

if (fc == '-') {
    if (mkobj(hi, TYPN, 2*cnt, 2)) return ERROR;
    for (z = INT_PTR(hi), si = INT_PTR(dst); 0 < ny--; si+=2) {
       if (si[0] != ERROR) {
          *z ++ = (qb) ? 89 - si[0] : si[0];
          *z ++ = (qb) ? 89 - si[1] : si[1];
           }
        }
    return sa_swap(dst, hi);
    }

if (fc == '+') {   /* mark my checks */
    mkobj(lo, TYPN, 3*cnt, 3);
    for (z = INT_PTR(lo), si = INT_PTR(dst); 0 < ny--; si+=2) {
        if (si[0] != ERROR) {
            *z ++ = (qb) ? 89 - si[0] : si[0];
            *z ++ = (qb) ? 89 - si[1] : si[1];
            for (j = 0;  j < B_SIZE; j++)
                xx[89-j] = (0 == map[j]) ? 0 : 8 ^ map[j];
            xx[89-si[1]] = xx[89-si[0]];
            xx[89-si[0]] = 0;
/* if (qt) xq_debug("xq.log", ny, xx); */
            *z++ =(0<xq_enprise(xx, ERROR));
            }
        }
    return sa_swap(dst, lo);
    }
return ERROR;
}

int     xq_rscan (map, x, dx)    /* rook moves */
int    *map;    /* board */
int     x;      /* position */
int     dx;     /* direction */
{
int     rc = 0; /* result */
int     cnt;

if (dx == 1)  cnt = NC - 1 - x % NC;
else if (dx == -1) cnt = x % NC;
else if (dx == NC) cnt = NR - 1 - x/NC;
else if (dx == -NC) cnt = x/NC;

for (rc = 0, x = x+dx ; 0<cnt--; x += dx) {
    rc++;
    if (map[x] > 0)  break;
    }
return  rc;
}

int     xq_cscan (map, x, dx, hit)      /* cannon moves */
int    *map;    /* board */
int     x;      /* position */
int     dx;     /* direction */
int    *hit;    /* target piece if any */
{
int     rc = 0; /* result */
int     cnt, i, k, mode;

if (dx == 1)  cnt = NC - 1 - x % NC;
else if (dx == -1) cnt = x % NC;
else if (dx == NC) cnt = NR - 1 - x/NC;
else if (dx == -NC) cnt = x/NC;

if (cnt == 0) {
    *hit = 0;
    return rc;
    }

for (mode = i = k = 0, x = x+dx ; i < cnt; i++, x += dx) {
    if (mode == 0 && map[x] > 0) {
        rc = i;
        mode = 1;
        }
    else if (mode && map[x] > 0) {
        k = i+1;
        break;
        }
    }
if (mode == 0) rc = cnt;
*hit = k;
return rc;
}

int xq_flip(map)    /* reverse board and swap red to black */
MATRIX map;
{
int     i, j, c;
int    *ix;

if (map == NULL || NULL  == map->ADDR) return ERROR;

if (MSIZE(map->ADDR) != B_SIZE) return ERROR;
if (MTYPE(map->ADDR) != TYPN) return ERROR;
ix = INT_PTR(map);

/* reverse by transpositions */

for (i = 0, j = B_SIZE-1; i<j; i++, j--) {
        c = ix[i];
        ix[i] = ix[j];
        ix[j] = c;
        }

for (i=0; i < B_SIZE; i++) ix[i] = (ix[i] == 0) ? 0 : ix[i] ^ 8;
return OK;
}

int     xq_vflip(src)   /* in situ board involution */
int    *src;
{
int     i, j, c;

/* reverse by transpositions */

for (i = 0, j = B_SIZE-1; i<j; i++, j--) {
        c = src[i];
        src[i] = src[j];
        src[j] = c;
        }

for (i=0; i < B_SIZE; i++) src[i] = (src[i] == 0) ? 0 : src[i] ^ 8;
return OK;
}

/*      Xianq Qi. 'En-prise' functin. */
/*      Is a white piece at position 'x' en prise */


int xq_shi(dst, lo, hi) /* Captures in Xiang qi position */
MATRIX  dst;            /* Table of numbers */
MATRIX  lo;             /* places or moves to test */
MATRIX  hi;             /* ASCII board. 90 characters */
{
int     xb[B_SIZE];
int    *map = INT_PTR(hi);
int    *ix, *iy;
int     n, np, nc;
int     i, j, p, q;

if (B_SIZE != MSIZE(hi->ADDR)) return ERROR;

if (lo == NULL ||
    0 == (n = MSIZE(lo->ADDR)) ||
    0 == (nc = MCOLS(lo->ADDR)))
    return mkik(dst, xq_enprise(map, ERROR));
if (nc > 2) return ERROR;
np = n/nc;
if (mkobj(dst, TYPN, np, 1))
    return ERROR;
ix = INT_PTR(lo);
iy = INT_PTR(dst);
if (nc == 1) {  /* list of points */
    iy = INT_PTR(dst);
    for (i = 0;  i < np; i++)   /* could avoid repeat tests */
        *iy++ = xq_enprise(map, *ix++);
    return OK;
    }
/*      case 2: check move table (p->q)[i]      */
for (i = 0; i < np; i++) {
    for (j = 0; j < B_SIZE; j++)
        xb[j] = map[j]; /* copy position */
    p = *ix++;
    q = *ix++;
    xb[q] = xb[p];      /* move p->q */
    xb[p] = 0;
    *iy++ = xq_enprise(xb, q);
    }
return OK;
}

int     xq_enprise(src, p)
int    *src;
int     p;
{
int     px, py;
int     j, jr, jc, k;
int     dzsq, block;
int     rc = 0;
int     rk = 0;         /* red king verify */

if (p == ERROR) {       /* find king */
    for (j = 0; j < B_SIZE && p == ERROR; j++) {
        if (src[j] == KING) p = j;
        }
    }
if (p == ERROR) return 1;       /* ILLEGAL. No king */
rk = (src[p] == KING);

px = p % 9;
py = p / 9;
for (j = jr = jc = 0; j < B_SIZE; j++) {
    dzsq = (px-jc)*(px-jc)+(py-jr)*(py-jr);
    switch (k = src[j]) {
    case B_R:   /* rook j->p */
        if (px == jc || py == jr)
            rc += (0 == rc_line(src, j, p));
        break;
    case B_C:   /* cannon j->p */
        if (px == jc || py == jr)
            rc += (1 == rc_line(src, j, p));
        break;
    case B_H:   /* horse  */
        if (dzsq == 5) {
            block = (9*I_SIGN(jr-py))+I_SIGN(jc-px);
            rc += src[p+block] == 0;
            }
        break;
    case B_P:   /* pawn */
        if (dzsq == 1) {
            if ((jr-py) == 1) rc++;
            if (jr < 5)
                rc += (jr == py);       /* sideways pawn move */
                }
        break;
    case B_A:   /* advisor */
        if (dzsq == 2)
            rc += A_BOX(p);
        break;
    case B_E:   /* elephant */
        if (dzsq == 8 && p > 44)
            rc += (src[(p+j)/2] == 0);
        break;
    case B_K:   /* king */
        if (dzsq == 1)
            rc += K_BOX(px, py);
        if (rk && px == jc)             /* king file check */
            rc += (0 == rc_line(src, j, p));
        break;
    default:
        break;
        }
     jc++;
     if (jc == 9) { jc = 0; jr++; }
     }
return rc;
}

int     rc_line(s, a, b) /* attack by rook or cannon */
int    *s;
int     a, b;
{
int    *t;
int     n  = I_ABS(b-a);
int     dx = I_SIGN (b-a);
int     rc = 0;

if (n>8) {
        n = n/9;
        dx = 9*dx;
        }
for (n--, t = s+a+dx; 0<n--; t+=dx)
    rc += (*t != 0);
return rc;
}

xq_debug(fn, key, pos)  /* dump a position */
char   *fn;
int     key;
int    *pos;
{
char   *xlate =".PAECHRK.paechrk";
int     i;
FILE   *op = fopen(fn,"aw");

fprintf(op, "\nposition %d\n", key);
for (i=0; i<B_SIZE;i++) {
    if (0 == i % 9) fputc('\n', op);
    fputc(xlate[pos[i]], op);
    }
fputc('\n', op);
fclose(op);
return  OK;
}

